<?php

namespace Drupal\micro_sso\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Routing\TrustedRedirectResponse;
use Drupal\micro_site\Entity\SiteInterface;
use Drupal\micro_site\SiteNegotiatorInterface;
use Drupal\user\UserInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\AccessDeniedHttpException;
use Symfony\Component\HttpFoundation\JsonResponse;
use Drupal\micro_sso\MicroSsoHelperInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;

/**
 * Default controller for the ucms_sso module.
 */
class MicroSsoController extends ControllerBase {

  /**
   * @var \Drupal\micro_site\SiteNegotiatorInterface
   */
  protected $negotiator;

  /**
   * @var \Drupal\micro_sso\MicroSsoHelperInterface
   */
  protected $microSsoHelper;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs the controller.
   *
   * @param \Drupal\micro_sso\MicroSsoHelperInterface $micro_sso_helper
   *   The micro sso helper service.
   * @param \Drupal\micro_site\SiteNegotiatorInterface $site_negotiator
   *   The site negotiator
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(MicroSsoHelperInterface $micro_sso_helper, SiteNegotiatorInterface $site_negotiator, EntityTypeManagerInterface $entity_type_manager) {
    $this->microSsoHelper = $micro_sso_helper;
    $this->negotiator = $site_negotiator;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('micro_sso.helper'),
      $container->get('micro_site.negotiator'),
      $container->get('entity_type.manager')
    );
  }

  public function callback(SiteInterface $site) {
    $url = $this->microSsoHelper->writeToken($site->getSiteUrl());
    unset($_GET['destination']);
    return new TrustedRedirectResponse($url);
  }

  /**
   * Check on the master host if the user is authenticated
   * and the request is valid.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   */
  public function check() {
    if (!$origin = $this->microSsoHelper->getOrigin()) {
      throw new AccessDeniedHttpException();
    }

    if (!$this->microSsoHelper->userIsAuthenticated()) {
      // we use a cacheable json response, has all really anonymous users
      // will end up there and this can be easily managed by the reverse
      // proxy cache this will be a 200 and not a 403, better chances of
      //  caching.
      // With cache_page activated we have a cache duration and the
      // Vary: Cookie header.
      // Note that we had an drupal_page_is_cacheable(FALSE); before, so the drupal cache
      // will not cache it, because the HTTP response contains 1 Access-Control-Allow-Origin
      // header which depends on the origin of the query, only Varnish can cache this anon
      // response using the Origin header as a cache key, not Drupal page cache
      return new JsonResponse(['status' => 401, 'reason' => 'User not authenticated on master']);
    }

    $site = $this->negotiator->loadByHostname($origin);
    if ($site instanceof SiteInterface) {
      // We check that the current user logged in on the master hosts is a
      // member of the factory site. Users with permission "administer site entities"
      // can always use the sso login on the site entities.
      if (!in_array($this->microSsoHelper->getCurrentUser()->id(), $site->getAllUsersId()) &&
      !$this->microSsoHelper->getCurrentUser()->hasPermission('administer site entities')) {
        return new JsonResponse(['status' => 401, 'reason' => 'User not allowed to login']);
      }
    }
    else {
      return new JsonResponse(['status' => 401, 'reason' => 'Origin not a micro site']);
    }

    // User 42 is logged on site A (master).
    // User 42 goes on site B (slave).
    // Site B does an AJAX request on A (this callback).
    //  - Token exists: it returns OK, then redirect the client on itself
    //    with session id as token on site B (the next callback).
    //  - Token is wrong: it returns NO, a JavaScript cookie is set and it
    //    prevents the user from loggin in for a few minutes.
    return new JsonResponse([
      'status' => 1,
      'uri' => $this->microSsoHelper->writeToken($origin),
    ]);
  }

  /**
   * Attempt to login the user on the micro site.
   *
   * @return \Symfony\Component\HttpFoundation\JsonResponse
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function login() {
    $request = $this->microSsoHelper->getRequest();
    $token = $request->query->get('s');

    if (empty($token)) {
      throw new AccessDeniedHttpException();
    }

    if (!preg_match('/^[a-zA-Z0-9_-]+$/', $token)) {
      throw new AccessDeniedHttpException();
    }

    // Retrieve stored temporary token storage
    $domain = $request->server->get('HTTP_HOST');
    $cid = $domain . ':' . $token;
    $destination = $request->query->get('destination');

    if ($this->microSsoHelper->userIsAuthenticated()) {
      // User is logged in, attempt token removal and exit.
      $this->microSsoHelper->getCacheSso()->delete($cid);
      return new JsonResponse(['status' => 401, 'reason' => 'User already authenticated']);
    }

    $entry = $this->microSsoHelper->getCacheSso()->get($cid);
    if (!$entry) {
      return new JsonResponse(['status' => 401, 'reason' => 'No corresponding entry to token found']);
    }
    $record = $entry->data;
    // ALWAYS invalidate this token for future use that's a ONE TIME thing.
    $this->microSsoHelper->getCacheSso()->delete($cid);

    // Check that this record was for this site, this IP, and with a good validity.
    if ($this->microSsoHelper->getRequest()->getClientIp() !== $record['ip'] || $domain != $record['origin'] || $record['validity'] < $this->microSsoHelper->getRequestTime()) {
      return new JsonResponse(['status' => 401, 'reason' => 'Token invalid']);
    }

    $uid = $record['uid'];
    $user = $this->entityTypeManager->getStorage('user')->load($uid);
    if (!$user instanceof UserInterface) {
      return new JsonResponse(['status' => 401, 'reason' => 'An error occurs with the user']);
    }

    user_login_finalize($user);
    return new JsonResponse(['success' => TRUE, 'destination' => $destination]);
  }

}
